/*----------------------------------------------------------------------------

  Copyright (c) 2015 rafael grompone von gioi <grompone@gmail.com>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>.

  ----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/*----------------------------------------------------------------------------*/
/* Fatal error, print a message to standard-error output and exit.            */
void error(char * msg)
{
  fprintf(stderr,"Error: %s\n",msg);
  exit(EXIT_FAILURE);
}

/*----------------------------------------------------------------------------*/
/* Memory allocation, print an error and exit if fail.                        */
void * xmalloc(size_t size)
{
  void * p;
  if( size == 0 ) error("xmalloc: zero size");
  p = malloc(size);
  if( p == NULL ) error("xmalloc: out of memory");
  return p;
}

/*----------------------------------------------------------------------------*/
/* Open file, print an error and exit if fail.                                */
FILE * xfopen(const char * path, const char * mode)
{
  FILE * f = fopen(path,mode);
  if( f == NULL ) error("xfopen: unable to open file");
  return f;
}

/*----------------------------------------------------------------------------*/
/* Close file, print an error and exit if fail.                               */
int xfclose(FILE * f)
{
  if( fclose(f) == EOF ) error("xfclose: unable to close file");
  return 0;
}

/*----------------------------------------------------------------------------*/
/* Read an ASC format image file.                                             */
double * read_asc(char * name, int * X, int * Y, int * Z, int * C)
{
  FILE * f;
  int i,n;
  double val;
  double * image;

  /* open file */
  f = xfopen(name,"rb"); /* open to read as a binary file (b option). otherwise,
                            in some systems, it may behave differently */

  /* read header */
  n = fscanf(f,"%d%*c%d%*c%d%*c%d",X,Y,Z,C);
  if( n!=4 || *X<=0 || *Y<=0 || *Z<=0 || *C<=0 ) error("invalid ASC file");

  /* get memory */
  image = (double *) xmalloc( *X * *Y * *Z * *C * sizeof(double) );

  /* read data */
  for(i=0; i<(*X * *Y * *Z * *C); i++)
    {
      n = fscanf(f,"%lf",&val);
      if( n!=1 ) error("invalid ASC file");
      image[i] = val;
    }

  /* close file */
  xfclose(f);

  return image;
}

/*----------------------------------------------------------------------------*/
/* Write an int image into a PGM file.                                        */
void write_pgm_image(int * image, int X, int Y, char * name)
{
  FILE * f;
  int x,y,n,v;

  /* check input */
  if( image == NULL || X < 1 || Y < 1 ) error("write_pgm_image: invalid image");

  /* open file */
  f = xfopen(name,"w");

  /* write header */
  fprintf(f,"P2\n");
  fprintf(f,"%d %d\n",X,Y);
  fprintf(f,"255\n");

  /* write data */
  for(n=0,y=0; y<Y; y++)
    for(x=0; x<X; x++)
      {
        v = image[x+y*X];
        if( v < 0 ) v = 0;
        if( v > 255 ) v = 255;
        fprintf(f,"%3d ",v);
        if(++n==16)  /* lines should not be longer than 70 characters  */
          {
            fprintf(f,"\n");
            n = 0;
          }
      }

  /* close file */
  xfclose(f);
}

/*----------------------------------------------------------------------------*/
/*                                    Main                                    */
/*----------------------------------------------------------------------------*/
int main(int argc, char ** argv)
{
  double alpha = 2.0;
  double min =   0.0;
  double max = 255.0;
  double * image_double;
  int    * image_int;
  int X,Y,Z,C,i,n;
  double m,M,mean,std;

  /* read input */
  if( argc < 3 )
    error("use: asc2norm_pgm in.asc out.pgm [alpha]\n"
          "     alpha: normalize to mean +/- alpha * std  (default alpha=2.0)");
  image_double = read_asc(argv[1], &X, &Y, &Z, &C);
  if( Z!=1 || C!=1 ) error("asc2norm_pgm only support 1 channel images.");
  if( argc == 4 )
    {
      alpha = atof(argv[3]);
      if( alpha <= 0.0 ) error("asc2norm_pgm: alpha must be positive");
    }

  /* compute mean and std */
  mean = 0.0;
  n = 0;
  for(i=0; i<X*Y*Z*C; i++)
    if( isfinite(image_double[i]) )
      {
        ++n;
        mean += image_double[i];
      }
  mean  /= (double) n;
  std = 0.0;
  for(i=0; i<X*Y*Z*C; i++)
    if( isfinite(image_double[i]) )
      std += (image_double[i] - mean) * (image_double[i] - mean);
  std = sqrt( std / (double) n );

  /* normalize image */
  m = (int) (mean - alpha * std);
  M = (int) (mean + alpha * std);
  if( M > m )
    for(i=0; i<X*Y*Z*C; i++)
      image_double[i] = min + (image_double[i]-m) * (max-min) / (M-m);
  else
    for(i=0; i<X*Y*Z*C; i++)
      image_double[i] = min;
  for(i=0; i<X*Y*Z*C; i++) if( image_double[i] < min ) image_double[i] = min;
  for(i=0; i<X*Y*Z*C; i++) if( image_double[i] > max ) image_double[i] = max;

  /* convertion to int */
  image_int = (int *) xmalloc( X * Y * Z * C * sizeof(int) );
  for(i=0; i<X*Y*Z*C; i++)
    image_int[i] = (int) image_double[i];

  /* write output */
  write_pgm_image(image_int, X, Y, argv[2]);

  /* free memory */
  free( (void *) image_double );
  free( (void *) image_int );

  return EXIT_SUCCESS;
}
/*----------------------------------------------------------------------------*/
