/*----------------------------------------------------------------------------

  Copyright (c) 2018-2020 rafael grompone von gioi <grompone@gmail.com>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  ----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "io.h"

/*----------------------------------------------------------------------------*/
#ifndef FALSE
#define FALSE 0
#endif /* !FALSE */

#ifndef TRUE
#define TRUE 1
#endif /* !TRUE */

/*----------------------------------------------------------------------------*/
/* PI */
#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif /* !M_PI */

/*----------------------------------------------------------------------------*/
/* Label for pixels with undefined gradient. */
#define NOTDEF -2.0

/*----------------------------------------------------------------------------*/
/* Fatal error, print a message to standard error and exit
 */
static void error(char * msg)
{
  fprintf(stderr,"error: %s\n", msg);
  exit(EXIT_FAILURE);
}

/*----------------------------------------------------------------------------*/
/* Memory allocation, print an error and exit if fail
 */
static void * xmalloc(size_t size)
{
  void * p;
  if( size == 0 ) error("xmalloc input: zero size");
  p = malloc(size);
  if( p == NULL ) error("out of memory");
  return p;
}

/*----------------------------------------------------------------------------*/
/* Normalized angle difference between 'a' and the symmetric of 'b'
   relative to a vertical axis.
 */
static double norm_angle(double a, double b)
{
  a -= b;
  while( a <= -M_PI ) a += 2.0*M_PI;
  while( a >   M_PI ) a -= 2.0*M_PI;

  return fabs(a) / M_PI;
}

/*----------------------------------------------------------------------------*/
/* Recursively add 4-connected pixels with a_error <= rho, to the region R,
   which is started with the seed pixel at (x0,y0).

   x0, y0:    coordinates of seed pixel
   a_error:   pointer to the normalized angle error map
   X, Y:      size of the image domain
   rho:       normalized error angle threshold
   R_x, R_y:  pointers to buffers to return coordinates of pixels in region R
   R_n:       pointer to return the size of region R

   returns the sum of the normalized angular error in region R

   It is assumed that R_x and R_y are correctly allocated
 */
static double grow_region( int x0, int y0, double * a_error, int X, int Y,
                           double rho, int * R_x, int * R_y, int * R_n )
{
  int neighbor_x[4] = {  0,  0, -1,  1 };
  int neighbor_y[4] = { -1,  1,  0,  0 };
  double sum;
  int n,k;

  /* check input */
  if( a_error == NULL || R_x == NULL || R_y == NULL || R_n == NULL ||
      X <= 0 || Y <= 0 || x0 <= 0 || x0 >= X-1 || y0 <= 0 || y0 >= Y-1 )
    error("grow_region: invalid input");

  /* initialize region */
  R_x[0] = x0;              /* add pixel to region */
  R_y[0] = y0;
  *R_n = 1;
  sum = a_error[x0+y0*X];   /* initialize error sum with error in seed pixel */
  a_error[x0+y0*X] = 1.0;   /* mark pixel as already used to not visit again */

  /* recursively add 4-neighbors of the region with a_error <= rho */
  for(n=0; n < *R_n; n++)
    for(k=0; k<4; k++)
      {
        int x = R_x[n] + neighbor_x[k];   /* try 4-connected neighbors */
        int y = R_y[n] + neighbor_y[k];

        /* a_error is only computed inside [1, X-2] x [1, Y-2] */
        if( x > 0 && x < X-1 && y > 0 && y < Y-1 && a_error[x+y*X] <= rho )
          {
            R_x[*R_n] = x;          /* add pixel to region R */
            R_y[*R_n] = y;
            ++(*R_n);               /* increase the region size */
            sum += a_error[x+y*X];  /* add angular error to the error sum */
            a_error[x+y*X] = 1.0;   /* mark pixel as already used */
          }
      }

  return sum;  /* return normalized error sum */
}

/*----------------------------------------------------------------------------*/
/* Grain Filter: remove 4-connected regions or visible or not visible pixels,
                 if the size is less than lambda.

   visible:  pointer to the output visibility masks
   X, Y:     size of the image domain
   N:        number of images
   lambda:   threshold on the size of regions to remove
   R_x, R_y: pointers to buffers to store coordinates of pixels in region R
             (this is just to avoid allocating new memory).

   R_x and R_y must be correctly allocated and its contents is destroyed.
 */
static void grain_filter( int * visible, int X, int Y, int N, int lambda,
                          int * R_x, int * R_y )
{
  int neighbor_x[4] = {  0,  0, -1,  1 };
  int neighbor_y[4] = { -1,  1,  0,  0 };
  int i, j, k, x, y, V, R_n;
  int * available;

  /* check input */
  if( visible == NULL || R_x == NULL || R_y == NULL )
    error("grain_filter: invalid input");

  /* get memory */
  available = (int *) xmalloc( X * Y * sizeof(int) );

  /* main loop */
  for(i=0; i<N; i++)
  for(V=0; V <= 255; V+=255) /* perform grain filter for 0 and 255 */
    {
      for(k=0; k<X*Y; k++) available[k] = TRUE;  /* initialize mask each loop */

      for(x=1; x<X-1; x++)
      for(y=1; y<Y-1; y++)
        if( available[x+y*X] && visible[x+y*X + i*X*Y] == V )
          {
            /* initialize region */
            available[x+y*X] = FALSE;
            R_x[0] = x;
            R_y[0] = y;
            R_n = 1;

            /* recursively add neighbors of the that match the criterion */
            for(k=0; k<R_n; k++)
            for(j=0; j<4; j++)
              {
                int xx = R_x[k] + neighbor_x[j];
                int yy = R_y[k] + neighbor_y[j];

                if( xx > 0 && xx < X-1 && yy > 0 && yy < Y-1 &&
                    available[xx+yy*X] && visible[xx+yy*X + i*X*Y] == V )
                  {
                    available[xx+yy*X] = FALSE;
                    R_x[R_n] = xx;
                    R_y[R_n] = yy;
                    ++R_n;
                  }
              }

            /* remove connected region if smaller than lambda */
            if( R_n < lambda )
              for(k=0; k<R_n; k++)
                visible[ R_x[k] + R_y[k]*X + i*X*Y ] = (V == 255) ? 0 : 255;
          }
    }

  /* free memory */
  free( (void *) available );
}

/*----------------------------------------------------------------------------*/
/* Computes visibility masks for a series registered images.

   images:   pointer to the image contents
   visible:  pointer to the output visibility masks
   X, Y:     size of the images domain
   N :       number of input images
   lambda:   threshold on the size of regions to remove in grain filter

   It is assumed that the output array 'visible' is correctly allocated.
 */
void visibility(double * images, int * visible, int X, int Y, int N, int lambda)
{
  double rho = 0.2;
  double * angles;     /* image gradient orientation angle, g_k in IPOL paper */
  double * a_error;    /* normalized angle error, \xi in IPOL paper */
  int * R_x;           /* region R in IPOL paper */
  int * R_y;
  int R_n;             /* size of region R */
  double logNT, logNFA;
  int i, j, k, x, y;

  /* check input */
  if( images == NULL || visible == NULL || X <= 0 || Y <= 0 || N < 2 )
    error("visibility: invalid input");

  /* get memory */
  angles  = (double *) xmalloc( X * Y * N * sizeof(double) );
  a_error = (double *) xmalloc( X * Y     * sizeof(double) );
  R_x     = (int *)    xmalloc( X * Y     * sizeof(int)    );
  R_y     = (int *)    xmalloc( X * Y     * sizeof(int)    );

  /* initialize output visibility maps */
  for(k=0; k<X*Y*N; k++) visible[k] = 255; /* initialize as not visible */

  /* compute gradient angles */
  for(i=0; i<N; i++)
  for(x=1; x<X-1; x++)
  for(y=1; y<Y-1; y++)
    {
      double dx = images[(x+1)+ y   *X + i*X*Y] - images[(x-1)+ y   *X + i*X*Y];
      double dy = images[ x   +(y+1)*X + i*X*Y] - images[ x   +(y-1)*X + i*X*Y];
      double mod = sqrt( dx * dx + dy * dy );

      /* note that atan2 handles correctly the case when ux is zero */
      if( mod <= 0.0 ) angles[x+y*X + i*X*Y] = NOTDEF;
      else             angles[x+y*X + i*X*Y] = atan2(dy, dx);
    }

  /* compare images to find matching patterns */
  for(i=0; i<N; i++)     /* loops on all pair of images i and j */
    for(j=i+1; j<N; j++)
      {
        /* compute normalized angle error */
        for(x=1; x<X-1; x++)
        for(y=1; y<Y-1; y++)
          {
            double a = angles[x+y*X + i*X*Y];
            double b = angles[x+y*X + j*X*Y];
            if( a == NOTDEF || b == NOTDEF ) a_error[x+y*X] = 1.0;
            else                             a_error[x+y*X] = norm_angle(a, b);
          }

        /* grow regions as potential match candidates */
        for(x=1; x<X-1; x++)
        for(y=1; y<Y-1; y++)
          if( a_error[x+y*X] <= rho )
            {
              /* compute region reg*/
              double s = grow_region(x, y, a_error, X, Y, rho, R_x, R_y, &R_n);

              /* The NFA can take very small values.  Computing log10(NFA)
                 instead of NFA is easier and reduce significantly the risk
                 of underflow due to the numerical representation of numbers.

                 The use of base 10 in the logarithm is arbitrary and it makes
                 no difference as the threshold epsilon=1 is always log(1)=0
                 regardless of the base.  The base ten has the advantage of
                 providing some intuition in the order of magnitude of the
                 NFA value, in case a value needs to be analyzed.  Indeed,
                 a log10(NFA) = -3 indicates that the NFA ~ 10^-3.
               */

              /* Number of test:
                 all pair of images ~ N^2                    ->  N^2
                 each pixel is the center of possible region ->  X*Y
                 we need to test regions of size 1 to X*Y    ->  X*Y
                 the number of polyominos of size n is about
                     0.316915 * 4.0625696^n / n              -> 0.3*4.06^n / n
               */
              logNT = 2.0 * log10(N) + 2.0 * log10(X) + 2.0 * log10(Y)
                    + log10(0.316915) + (double) R_n * log10(4.0625696)
                    - log10( (double) R_n );

              /* NFA <= NT * s^n / n!

                 log(n!) is bounded by Stirling's approximation:
                   n! >= sqrt(2pi) * n^(n+0.5) * exp(-n)
                 then, log10(NFA) <= log10(NT) + n*log10(s)
                                                 - log10(latter expansion)
               */
              logNFA = logNT + (double) R_n * log10(s)
                     - 0.5 * log10(2.0 * M_PI)
                     - ( (double) R_n + 0.5 ) * log10( (double) R_n )
                     + (double) R_n * log10(exp(1.0));

              /* if a meaningful match is found,
                 mark the region as visible in both images */
              if( logNFA < 0.0 )  /* NFA < 1  <=> log(NFA) < log(1) = 0.0 */
                for(k=0; k<R_n; k++)
                  {
                    int xx = R_x[k];
                    int yy = R_y[k];

                    visible[xx+yy*X + i*X*Y] = 0; /* image A as visible */
                    visible[xx+yy*X + j*X*Y] = 0; /* image B as visible */
                  }
            }
      }

  /* grain filter: remove connected regions of less than lambda pixels
     (the same region memory R_x and R_y are used as they are no longer used )*/
  grain_filter(visible, X, Y, N, lambda, R_x, R_y);

  /* correct border pixels because the visibility masks are only computed
     inside [1, X-2] x [1, Y-2]: set values at image border to the same value
     as the pixel just one pixel inside the image domain */
  for(i=0; i<N; i++)
    {
      for(x=0;x<X;x++) visible[ x + 0   *X+i*X*Y] = visible[ x + 1   *X+i*X*Y];
      for(x=0;x<X;x++) visible[ x +(Y-1)*X+i*X*Y] = visible[ x +(Y-2)*X+i*X*Y];
      for(y=0;y<Y;y++) visible[ 0 + y   *X+i*X*Y] = visible[ 1 + y   *X+i*X*Y];
      for(y=0;y<Y;y++) visible[X-1+ y   *X+i*X*Y] = visible[X-2+ y   *X+i*X*Y];
    }

  /* free memory */
  free( (void *) angles);
  free( (void *) a_error );
  free( (void *) R_x );
  free( (void *) R_y );
}

/*----------------------------------------------------------------------------*/
/*                                    main                                    */
/*----------------------------------------------------------------------------*/
int main(int argc, char ** argv)
{
  double * images;  /* input images */
  int * visible;    /* output visibility masks, V_k in IPOL paper */
  int lambda;       /* parameter: size of regions to remove in grain filter */
  int X, Y, N, XX, YY, n, k;

  /* usage */
  if( argc < 4 )
    error("usage: visibility <lambda> image1 image2 [image3 ... imageN]\n"
          "       lambda: regions of less than lambda pixels would be removed\n"
          "       images: only PGM or ASC file formats are handled");

  /* read input */
  lambda = atoi(argv[1]);  /* read parameter lambda */
  N = argc - 2;
  for(n=0; n<N; n++)
    {
      /* read new image */
      double * image = read_image(argv[n+2], &XX, &YY);

      if( XX < 1 || YY < 1 ) error("invalid image size");
      if( n>0 && ( XX!=X || YY!=Y ) ) error("images must have the same size");

      /* get memory */
      if( n == 0 )
        {
          X = XX;
          Y = YY;
          images   = (double *) xmalloc( X * Y * N * sizeof(double) );
          visible  = (int *)    xmalloc( X * Y * N * sizeof(int)    );
        }

      /* copy new image to buffer 'images' */
      for(k=0; k<X*Y; k++) images[k + n*X*Y] = image[k];

      free( (void *) image );
    }

  /* call visibility detector */
  visibility(images, visible, X, Y, N, lambda);

  /* write output cloud masks */
  for(n=0; n<N; n++)
    {
      char filename[512];
      sprintf(filename, "%s_visibility.pgm", argv[n+2]);
      write_pgm_image(visible + n*X*Y, X, Y, filename);
    }

  /* free memory */
  free( (void *) images  );
  free( (void *) visible );

  return EXIT_SUCCESS;
}
/*----------------------------------------------------------------------------*/
